{%- extends "!layout.html" %}
{%- block breadcrumbs %}
{{ super() }}
<!--REGION_DISCLAIMER_DO_NOT_REMOVE-->
{%- endblock %}


{%- block footer %}



<div class="footer">
    <script type="text/javascript" src="{{ pathto('_static/shortbreadv1.js',1 ) }}"></script>
    <script type="text/javascript">
            const shortbread = AWSCShortbread({
                domain: ".amazonaws.com",
});
            shortbread.checkForCookieConsent();
    </script>
    <a href="http://aws.amazon.com/privacy">Privacy</a> | <a href="http://aws.amazon.com/terms">Site Terms</a> | <a
        href="#" onclick="shortbread.customizeCookies();">Cookie preferences</a> |
    &copy; Copyright {{ copyright }} <a href="http://sphinx.pocoo.org/">Sphinx</a>.
</div>

{%- endblock %}
